__all__ = [
    "BATCH_SIZE",
    "FEATURE",
    "SEQUENCE",
    "SEQ_LEN",
    "TARGET",
    "TIME",
    "VALUE",
]

BATCH_SIZE = "batch_size"
FEATURE = "feature"
SEQUENCE = "sequence"
SEQ_LEN = "seq_len"
TARGET = "target"
TIME = "time"
VALUE = "value"
